#include <ssl/bio.h>
#include <ssl/err.h>
#include <ssl/ssl.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

enum {
	MODE_SERVER,
	MODE_CLIENT
};

enum {
	BUF_LEN = 512
};

#define CERTIFICATE "certs/certificate.pem"

void server_mode (const char *host_insecure, const char *host_secure);
void client_mode (const char *host_insecure, const char *host_secure);

static SSL_CTX *ossl_ctx = NULL;

int main (int argc, *char argv[])
{
	int mode, port, sport;
	const char *host;
	
	if (argc != 4) {
		printf("Simple OpenSSH tunnelling program.\n");
		printf("Usage: tst client|server <insecure_host:port> <secure_host:port>\n");
		exit(0);
	}

	if (strcmp(argv[1], "client") == 0) {
		mode = MODE_CLIENT;	
	} else if (strcmp(argv[1], "server") == 0) {
		mode = MODE_SERVER;
	} else {
		printf("Error: You did not specify an operation mode.\n");
		exit(2);
	}
	
	if (ossl_ctx == NULL) {
		SSL_library_init();
		OpenSSL_add_all_algorithms();
		ossl_ctx = SSL_CTX_new(SSLv3_client_method());
	}

	switch (mode)
	{
	default:
	case (MODE_SERVER):
		client_mode(argv[2], argv[3]);
		break;
		
	case (MODE_CLIENT):
		client_mode(argv[2], argv[3]);
		break;
	}
	
	SSL_CTX_free(ossl_ctx);
	
	return 0;
}

void server_mode (const char *host_insecure, const char *host_secure)
{
	
}

void client_mode (const char *host_insecure, const char *host_secure)
{
	BIO *bio_secure, *bio_insecure;
	SSL *ssl;
	
	fd_set fdset, fdnull;
	char buffer[BUF_LEN];
	int len, fd_secure, fd_insecure, maxfd;
	
	/* Set up the secure connection */
	bio_secure = BIO_new_ssl_connect(ossl_ctx);
	BIO_get_ssl(bio_secure, &ssl);
	SSL_set_mode(ssl, SSL_MODE_AUTO_RETRY);
	
	BIO_set_conn_hostname(bio_secure, host_secure);
	
	if (BIO_do_connect(bio_secure) <= 0) {
		fprintf(stderr, "Connection failed, unable to connect secure socket.\n");
		return;
	}
	
	/* Set up the insecure connection, the local port */
	bio_insecure = BIO_new_connect(host_insecure);
	if (bio_insecure == NULL) {
		fprintf(stderr, "d");
		return;
	}
	
	if (BIO_do_connect(bio_insecure) <= 0) {
		fprintf(stderr, "Connection failed, unable to connect insecure socket.\n");
		return;
	}
	
	fd_secure = BIO_get_fd(bio_secure);
	fd_insecure = BIO_get_fd(bio_insecure);
	
	if (fd_secure > fd_insecure)
		maxfd = fd_secure + 1;
	else
		maxfd = fd_insecure + 1;
	
	/* Set up the fd_set structures to monitor our sockets. */
	FD_ZERO(&fdset);
	FD_ZERO(&fdnull);
	FD_SET(fd_secure, &fdset);
	FD_SET(fd_insecure, &fdset);
	
	memset(buffer, '\0', sizeof(char) * BUF_LEN);
	len = 0;
	
	while (1) {
		/* Monitor sockets, resume execution when data is available. */
		printf("Entering sleep state... ");
		select(maxfd, &fdset, &fdnull, &fdnull, NULL);
		printf("reawakended, data access...\n");
		
		len = BIO_read(bio_insecure, buffer, BUF_LEN -1);
		if (len > 0)
			BIO_write(bio_secure, buffer, len);
		
		len = BIO_read(bio_secure, buffer, BUF_LEN -1);
		if (len > 0)
			BIO_write(bio_insecure, buffer, len);
	}
	
	BIO_free_all(bio_secure);
	BIO_free_all(bio_insecure);
}
